﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Globalization;
using System.Net;
using System.IO;
using System.Text.RegularExpressions;
using MulticastTV;
using DirectShowLib;
using UdpSrcCom;
using Player;

namespace MctvPlaylistEditor
{
    
    public partial class Form1 : Form
    {
        static string UserAppPath;
        static string StartupPath;
        static string language = "default";
        static ImageList il;
        static bool RenamingTipShowed = false;
        static List<string> groupList = new List<string>();
        static bool delayedTreeConstruction = false;

        public Form1()
        {
            InitializeComponent();

            Version ver = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version;
            this.Text += " " + ver.Major + "." + ver.Minor;

            il = new ImageList();
            il.Images.Add(Properties.Resources.Default);
            il.Images.Add(Properties.Resources.Green);
            il.Images.Add(Properties.Resources.film);    //film    2
            il.Images.Add(Properties.Resources.film_go);

            treeView1.ImageList = il;
            treeView1.ContextMenuStrip = contextMenuStrip2;

            UserAppPath = System.Windows.Forms.Application.UserAppDataPath;
            StartupPath = Application.StartupPath;
            setLanguage();
            dataGridView1.Columns[0].DataPropertyName = "Ip";
            dataGridView1.Columns[1].DataPropertyName = "Chan";
            dataGridView1.Columns[2].DataPropertyName = "Name";
            dataGridView1.Columns[3].DataPropertyName = "Group";
            dataGridView1.Columns[4].DataPropertyName = "Audio";
            dataGridView1.Columns[5].DataPropertyName = "EpgId";
            bindingSource1.DataSource = ChannelTable.menu.Tables["Menu"];
            dataGridView1.DataSource = bindingSource1;

            dataGridView1.Sort(dataGridView1.Columns[1], ListSortDirection.Ascending);

            toolTip1.SetToolTip(toolStripTextBox1.Control, "Enter ip address prefix. M3u format only.");

            m_ComboBoxCollection = new List<ComboBoxItem>();        // Populate combobox with names of network interfaces.

            List<string> list = m_nics.getNames();
            foreach (string name in list)
                m_ComboBoxCollection.Add(new ComboBoxItem(name, m_nics.getIndexFromName(name)));

            comboBox1.DataSource = m_ComboBoxCollection;
            comboBox1.DisplayMember = "Name";
            comboBox1.ValueMember = "Index";

            // Load settings from last time
            if (Properties.Settings.Default.WaitSdt)
                checkBox1.CheckState = CheckState.Checked;
            else
                checkBox1.CheckState = CheckState.Unchecked;

            numericUpDown2.Value = Properties.Settings.Default.WaitTimeout;
            numericUpDown3.Value = Properties.Settings.Default.NumberOfThreads;
            numericUpDown4.Value = Properties.Settings.Default.Port;
            textBox1.Text = Properties.Settings.Default.StartIp;
            textBox2.Text = Properties.Settings.Default.StopIp;

            int ni = Properties.Settings.Default.NicIndex;
            if (comboBox1.Items.Count > Properties.Settings.Default.NicIndex)
                comboBox1.SelectedIndex = ni;

            toolStripComboBox1.SelectedIndex = Properties.Settings.Default.XspfExtension;
            

        }

        //=====================================================================================================
        // Editor
        //=====================================================================================================

        private void openPlaylistToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openFileDialog1.CheckFileExists = true;
            openFileDialog1.Filter = "atvlist, m3u, xls, csv, xspf, js (*.atvlist;*.m3u;*.xls;*.csv;*.xspf)|*.atvlist;*.m3u;*.xls;*.csv;*.xspf;*.js";
            openFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                fillChannelTable(openFileDialog1.FileName);
                constructTreeFromTable();
            }
        }

        private void m3uToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "m3u (*.m3u)|*.m3u";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                  ExportToM3U.Start(saveFileDialog1.FileName,true,toolStripTextBox1.Text);
            }
        }

        private void m3uwithoutChanNumbersToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "m3u (*.m3u)|*.m3u";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToM3U.Start(saveFileDialog1.FileName, false, toolStripTextBox1.Text);
            }
        }

        private void m3uTanoPlayerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "m3u (*.m3u)|*.m3u";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToM3uTano tp = new ExportToM3uTano();
                tp.Start(saveFileDialog1.FileName, "");
            }
        }

        private void xspfToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "xspf (*.xspf)|*.xspf";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToXspf.Start(saveFileDialog1.FileName, toolStripTextBox1.Text,toolStripComboBox1.SelectedIndex);
            }
        }

        private void atvlistToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "atvlist(*.atvlist)|*.atvlist";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToAtvlist.Start(saveFileDialog1.FileName);
            }
        }

        private void csvToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "csv (*.csv)|*.csv";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToCSV csv = new ExportToCSV();
                csv.Start(saveFileDialog1.FileName);
            }
        }

        private void simplicitySTBGUIToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "js (*.js)|*.js";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToSimplicityJs.Start(saveFileDialog1.FileName,false);
            }
        }

        private void multicastTVRecordingServerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "js (*.js)|*.js";
            saveFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportToSimplicityJs.Start(saveFileDialog1.FileName,true);
            }
        }

        private void groupListToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "xml (*.xml)|*.xml";
            saveFileDialog1.FileName = @"groups_" + Thread.CurrentThread.CurrentCulture.IetfLanguageTag + "_utf8.xml";
            openFileDialog1.InitialDirectory = StartupPath + @"\parse\";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportGroups eg = new ExportGroups();
                eg.Start(saveFileDialog1.FileName);
            }
        }

        private void add1ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(1);
        }

        private void subtract1ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(-1);
        }

        private void add10ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(10);
        }

        private void subtract10ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(-10);
        }

        private void add100ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(100);
        }

        private void subtract100ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            addNum(-100);
        }

        private void moveUPToolStripMenuItem_Click(object sender, EventArgs e)
        {
            swapWithRow(false);
        }

        private void moveDownToolStripMenuItem_Click(object sender, EventArgs e)
        {
            swapWithRow(true);
        }

        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                foreach (DataGridViewRow drw in dataGridView1.SelectedRows)
                {
                    if (drw.IsNewRow != true)
                        dataGridView1.Rows.Remove(drw);
                }
                constructTreeFromTable();
            }
            catch { }
        }

        private void dataGridView1_RowHeaderMouseClick(object sender, DataGridViewCellMouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right && e.RowIndex >= 0)
            {
                if (dataGridView1.SelectedRows.Count <= 1)
                {
                    dataGridView1.Rows[e.RowIndex].Selected = true;
                    dataGridView1.CurrentCell = dataGridView1.Rows[e.RowIndex].Cells[1];
                }
                contextMenuStrip1.Show(Cursor.Position);
            }
        }

        private void dataGridView1_CellValidating(object sender, DataGridViewCellValidatingEventArgs e)
        {
            dataGridView1.Rows[e.RowIndex].ErrorText = "";
            int newInteger;

            // Don't try to validate the 'new row' until finished 
            // editing since there
            // is not any point in validating its initial value.
            if (dataGridView1.Rows[e.RowIndex].IsNewRow) { return; }

            if (e.ColumnIndex == 1)
            {
                if (!int.TryParse(e.FormattedValue.ToString(),
                    out newInteger) || newInteger < 0)
                {
                    MessageBox.Show("Value has to be an integer.");
                    e.Cancel = true;
                }
            }
            else
            {
                if (e.FormattedValue.ToString() == "")
                {
                    DataRowView dr = (DataRowView)dataGridView1.Rows[e.RowIndex].DataBoundItem;
                    dr.Row[e.ColumnIndex] = "";
                }
                
            }
        }

        private void dataGridView1_DataError(object sender, DataGridViewDataErrorEventArgs e)
        {
            if (e.Context == DataGridViewDataErrorContexts.Commit)
            {
                // We just assume an ip duplicate !!!
                MessageBox.Show("Error. Check if ip address allready exists.");
            }
            if (e.Context == DataGridViewDataErrorContexts.CurrentCellChange)
            {

            }
            if (e.Context == DataGridViewDataErrorContexts.Parsing)
            {

            }
            if (e.Context == DataGridViewDataErrorContexts.LeaveControl)
            {

            }
        }

        private void selectAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            dataGridView1.SelectAll();
        }

        private void loadEpgStringsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openFileDialog1.CheckFileExists = true;
            openFileDialog1.Filter = "txt (*.txt)|*.TXT";
            openFileDialog1.FileName = "";
            //openFileDialog1.InitialDirectory = StartupPath + @"\import\";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                loadEpgStrings(openFileDialog1.FileName);
            }
        }

        private void dataGridView1_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
                try
                {
                    foreach (DataGridViewRow drw in dataGridView1.SelectedRows)
                    {
                        if (drw.IsNewRow != true)
                            dataGridView1.Rows.Remove(drw);
                    }
                    constructTreeFromTable();
                }
                catch { }

            // move up down
            if (e.KeyCode == Keys.Down && dataGridView1.SelectedRows.Count > 0)
            {
                swapWithRow(true);
                e.Handled = true;
            }
            if (e.KeyCode == Keys.Up && dataGridView1.SelectedRows.Count > 0)
            {
                swapWithRow(false);
                e.Handled = true;
            }
        }

        /// <summary>
        /// Create internal channel information Datatable from file
        /// </summary>
        /// <param name="file">Input file</param>
        static public void fillChannelTable(string file)
        {
            char[] plusdelimiter = { '+' };

            //Adittional informationm in case of m3u playlist.
            bool aditional = false;

            // Datatable each line one channel (multiple audio) ATVLIST IMPORT
            DataTable channels = Import.array(StartupPath, UserAppPath, file, ref aditional);
            if (channels == null)
            {
                return;
            }

            ChannelTable.menu.Tables["Menu"].Clear();
            DataTable groupst = new DataTable();
            bool failed = false;

            if (aditional)
            {
                //Read xml
                groupst.Columns.Add("Name", typeof(string));
                groupst.Columns.Add("Group", typeof(string));
                groupst.Columns.Add("Audio", typeof(string));
                groupst.Columns["Name"].DefaultValue = "";
                groupst.Columns["Group"].DefaultValue = "";
                groupst.Columns["Audio"].DefaultValue = "";
                groupst.TableName = Thread.CurrentThread.CurrentCulture.IetfLanguageTag;
                groupst.PrimaryKey = new DataColumn[] { groupst.Columns["Name"] };
                try
                {
                    string xmlgroups = StartupPath + @"\parse\groups_" + Thread.CurrentThread.CurrentCulture.IetfLanguageTag + "_utf8.xml";
                    if (File.Exists(xmlgroups))
                        groupst.ReadXml(xmlgroups);
                    else
                        groupst.ReadXml(StartupPath + @"\parse\groups_default_utf8.xml");
                }
                catch { failed = true; }
            }

            foreach (DataRow dr in channels.Rows)
            {
                int i = 0;
                foreach (string str in (string[])dr["Ip"])
                {
                    DataRow menurow;
                    menurow = ChannelTable.menu.Tables["Menu"].NewRow();
                    menurow[0] = str.Split(plusdelimiter)[1];
                    menurow[1] = dr["Channel"];
                    menurow[2] = dr["Name"];
                    menurow[3] = dr["Group"];
                    menurow[4] = str.Split(plusdelimiter)[0];
                    menurow[5] = dr["EpgId"];

                    if (aditional == true && !failed && groupst.Rows.Count > 0)
                    {
                        // Here we get audio and language information if we imported from m3u.
                        DataRow tdt = groupst.Rows.Find(dr["Name"]);
                        if (tdt != null)
                        {
                            menurow[3] = tdt["Group"];
                            menurow[4] = tdt["Audio"];
                        }
                    }
  
                    bool exists = false;
                    //check if input is unique
                    foreach (DataRow drc in ChannelTable.menu.Tables["Menu"].Rows)
                    {
                        if (String.Equals(drc[0].ToString(), menurow[0].ToString()))
                        {
                            exists = true;
                        }
                    }
                    if (!exists)
                        ChannelTable.menu.Tables["Menu"].Rows.Add(menurow);

                    i++;
                }
            }

 
           //Save this INTERNAL datatable to a file
           // ChannelTable.menu.Tables["Menu"].WriteXml(UserAppPath + @"\..\channellist.xml");
        }


        // FIXME: To slow!!
        private void addNum(int n)
        {
            if (dataGridView1.SortedColumn != dataGridView1.Columns[1])
                dataGridView1.Sort(dataGridView1.Columns[1], ListSortDirection.Ascending);

            // all from selected on
            if (dataGridView1.SelectedRows.Count <= 1)
            {
                DataGridViewCell currentcell = dataGridView1.CurrentCell;
                int start = dataGridView1.CurrentCell.RowIndex;
                int stop = dataGridView1.Rows.Count - 1;

                ChannelTable.menu.Tables["Menu"].BeginLoadData();
                List<DataRowView> after = new List<DataRowView>();
                for (int i = start; i < stop; i++)
                {
                    DataRowView dwt = (DataRowView)dataGridView1.Rows[i].DataBoundItem;
                    after.Add(dwt);
                }
                foreach (DataRowView dwt in after)
                {
                    dwt.Row[1] = Math.Max(0, (int)dwt.Row[1] + n);
                }
                ChannelTable.menu.Tables["Menu"].EndLoadData();
               
                if (currentcell.RowIndex>=0)
                    dataGridView1.CurrentCell = currentcell;

            }
            // only selected
            else if (dataGridView1.SelectedRows.Count > 1)
            {
                int[] indexes = { 0 };
                int i = 0;
                ChannelTable.menu.Tables["Menu"].BeginLoadData();
                foreach (DataGridViewRow dw in dataGridView1.SelectedRows)
                {
                    indexes[i] = dw.Index;
                    DataRowView dwt = (DataRowView)dw.DataBoundItem;
                    if (dwt != null && !dwt.IsNew)
                        dwt.Row[1] = Math.Max(0, (int)dwt.Row[1] + n);
                }
                ChannelTable.menu.Tables["Menu"].EndLoadData();
            }

            constructTreeFromTable();
        }

        // FIXME !!!
        private void swapWithRow(bool up)
        {
            if (dataGridView1.SortedColumn != dataGridView1.Columns[1])
                dataGridView1.Sort(dataGridView1.Columns[1], ListSortDirection.Ascending);

            if (dataGridView1.SelectedRows.Count > 0)
            {
                DataGridViewRow current = dataGridView1.SelectedRows[0];
                int i = 0;
                if (up && current.Index + 1 < dataGridView1.Rows.Count - 1)
                    i = 1;
                else if (!up && current.Index - 1 >= 0)
                    i = -1;

                DataGridViewRow next = dataGridView1.Rows[current.Index + i];
                int curChan = (int)current.Cells[1].Value;
                int nextChan = (int)next.Cells[1].Value;
                //int diff = Math.Abs(curChan - nextChan);
                //if (diff > 1)
                //{
                //    if (up)
                //        nextChan = curChan + 1;
                //    else
                //        curChan = nextChan + 1;
                //}

                // swap it
                if ((up && ((nextChan > curChan) && !current.Equals(next))) || (!up && ((nextChan < curChan) && !current.Equals(next))))
                {
                    ChannelTable.menu.Tables["Menu"].BeginLoadData();
                    ((DataRowView)current.DataBoundItem)[1] = nextChan;
                    ((DataRowView)next.DataBoundItem)[1] = curChan;
                    ChannelTable.menu.Tables["Menu"].EndLoadData();

                    dataGridView1.CurrentCell = dataGridView1.Rows[dataGridView1.Rows.IndexOf(next)].Cells[1];
                    next.Selected = true;
                }
            }

            delayedTreeConstruction = true;
        }

        private void loadEpgStrings(string file)
        {
            //clear current strings
            foreach (DataRow dr in ChannelTable.menu.Tables["Menu"].Rows)
            {
                    dr["EpgId"] = "";
            }

            StreamReader sr = new StreamReader(file, Encoding.UTF8);
            char[] eq = { '=' };
            string line = sr.ReadLine();
            while (line != null)
            {
                if (!line.Contains(""))
                    continue;

                string[] linedata = line.Split(eq);
                if (linedata.Length < 2)
                    continue;

                string id = linedata[1];
                string name = linedata[0];
               
                foreach (DataRow dr in ChannelTable.menu.Tables["Menu"].Rows )
                {
                    if (dr["Name"].ToString().ToUpper() == name.ToUpper())
                    {
                        dr["EpgId"] = id;
                        break;
                    }
                }
                line = sr.ReadLine();
            }
            sr.Close();
        }


        private void constructTreeFromTable()
        {
            treeView1.Nodes.Clear();
            groupList.Clear();
            
            treeView1.BeginUpdate();

            TreeNode tn = new TreeNode();
            tn.Name = "";
            tn.Text = "No Group";
            tn.ImageIndex = 0;
            treeView1.Nodes.Add(tn);

            //if (dataGridView1.SortedColumn != dataGridView1.Columns[1])
            //{
            //    dataGridView1.Sort(dataGridView1.Columns[1], ListSortDirection.Descending);
            //}

            foreach (DataGridViewRow dr in dataGridView1.Rows)
            {
                if (dr.IsNewRow)
                    continue;

                string[] groups = ((string)dr.Cells[3].Value).Split(",".ToCharArray());

                foreach (string group in groups)
                {
                    // is group allready in the tree
                    TreeNode[] tngrps;
                    tngrps = treeView1.Nodes.Find(group, false);

                    TreeNode tngrp;

                    if (tngrps.Length == 0)
                    {
                        // add root node
                        TreeNode tnn = new TreeNode();
                        tnn.Text = group;
                        tnn.Name = group;
                        tnn.ImageIndex = 1;
                        tnn.ToolTipText = "Click to rename";
                           treeView1.Nodes.Add(tnn);
                           tnn.ContextMenuStrip = contextMenuStrip3;
                        tngrp = tnn;

                        groupList.Add(group);
                    }
                    else
                    {
                        tngrp = tngrps[0];
                    }

                    // add child
                    TreeNode tnc = new TreeNode();
                    tnc.Text = ((int)dr.Cells[1].Value).ToString() + ". " + (string)dr.Cells[2].Value;
                    tnc.Name = (string)dr.Cells[0].Value; // Name is ip
                    tnc.Tag = new ChanInfo((int)dr.Cells[1].Value, (string)dr.Cells[3].Value, (string)dr.Cells[0].Value, (string)dr.Cells[2].Value); // All groups that this channel belongs.
                    tnc.ImageIndex = 2;
                    tnc.SelectedImageIndex = 3;
                    tnc.ContextMenuStrip = contextMenuStrip4;
                    tngrp.Nodes.Add(tnc);
                }
            }

            treeView1.EndUpdate();
            UpdateContextMenu();
        }

        private void dataGridView1_KeyUp(object sender, KeyEventArgs e)
        {
            if (delayedTreeConstruction)
            {
                delayedTreeConstruction = false;
                constructTreeFromTable();
            }
        }

        private void updateTableFromTree(List<TreeNode> tnc)
        {
            foreach (TreeNode tn in tnc)
            { 
                ChanInfo ci = (ChanInfo)tn.Tag;
                DataRow dr = ChannelTable.menu.Tables["Menu"].Rows.Find(ci.Ip);
                if (dr != null)
                {
                    dr["Group"] = ci.Group;
                    dr["Chan"] = ci.ChanNum;
                }
            }
        }

        private void updateTableFromTree(TreeNodeCollection tnc)
        {
            foreach (TreeNode tn in tnc)
            {
                ChanInfo ci = (ChanInfo)tn.Tag;
                DataRow dr = ChannelTable.menu.Tables["Menu"].Rows.Find(ci.Ip);
                if (dr != null)
                {
                    dr["Group"] = ci.Group;
                    dr["Chan"] = ci.ChanNum;
                }
            }
        }

        TreeNode tnToMove=null;

        private void treeView1_ItemDrag(object sender, ItemDragEventArgs e)
        {
            tnToMove = (TreeNode)e.Item;
            DoDragDrop(e.Item, DragDropEffects.Move | DragDropEffects.Copy);
        }

        private void treeView1_DragDrop(object sender, DragEventArgs e)
        {
            // Find out if destination is in the same group.
            // If that is true we just swap channel numbers,
            // else we move to different group.
            if (tnToMove.Level == 0)
                return;

           bool copy =false;
           if ((e.KeyState & 8) == 8 && (e.AllowedEffect & DragDropEffects.Copy) == DragDropEffects.Copy)
               copy = true;
           
            TreeNode destnode = treeView1.GetNodeAt(treeView1.PointToClient(new Point(e.X,e.Y)));
            MoveItem(copy,destnode);
    
        }

        private void MoveItem(bool copy, TreeNode destnode)
        { 
            TreeNode newnode = (TreeNode)tnToMove.Clone();

            if (destnode != null && tnToMove != null )
            {
               TreeNode destParentNode;
                bool multiplegroups = false;
                if (((ChanInfo)newnode.Tag).Group.Contains(","))
                    multiplegroups = true;

                if (destnode.Level == 1)
                    destParentNode = destnode.Parent;
                else
                    destParentNode = destnode;

                if (destParentNode != tnToMove.Parent) // Copy or move between root nodes.
                {
                    ChanInfo ci = (ChanInfo)newnode.Tag;
                    bool duplicate = false;
                    int indextoinsert = findIndexToInsert(destParentNode, ci.ChanNum, ci.Ip, ref duplicate);

                    if (!duplicate)
                    {
                        destParentNode.Nodes.Insert(indextoinsert, newnode);
                        UpdateBelongingGroups(ci, tnToMove.Parent.Name, destParentNode.Name, copy);

                        ((ChanInfo)tnToMove.Tag).Group = ci.Group;
                        if (!copy || destParentNode.Name == "")
                        {
                            tnToMove.Remove();
                            tnToMove = null;
                        }
                        List<TreeNode> tnlist = new List<TreeNode>();
                        tnlist.Add(newnode);
                        updateTableFromTree(tnlist);
                        if (destParentNode.Name == "" && multiplegroups)
                            constructTreeFromTable(); // Lazy...
                    }
                }
                else //Swap channel numbers
                {
                    if (destnode.Level < 1)
                        return;
                    
                    ChanInfo currentDest = (ChanInfo)destnode.Tag;
                    ChanInfo currentSource = (ChanInfo)tnToMove.Tag;
                    
                    int currentDestNum = currentDest.ChanNum;
                    
                    // Update Channel Info
                    currentDest.ChanNum = currentSource.ChanNum;
                    currentSource.ChanNum = currentDestNum;

                    // Update Text
                    destnode.Tag = currentSource;
                    destnode.Text = currentSource.ChanNum.ToString() + ". " + currentSource.Name;
                    tnToMove.Tag = currentDest;
                    tnToMove.Text = currentDest.ChanNum.ToString() + ". " + currentDest.Name;

                    List<TreeNode> tnlist = new List<TreeNode>();
                    tnlist.Add(destnode);
                    tnlist.Add(tnToMove);
                    updateTableFromTree(tnlist);
                
                }
            }
        }

        private int findIndexToInsert(TreeNode parent, int chanNum, string ip, ref bool duplicate)
        {
            int index = 0;
            foreach (TreeNode tn in parent.Nodes)
            {
                if (((ChanInfo)tn.Tag).Ip == ip)
                    duplicate = true;
                
                if (((ChanInfo)tn.Tag).ChanNum >= chanNum)
                    break;
                index += 1;
            }
            return index;
        }

        private void treeView1_DragEnter(object sender, DragEventArgs e)
        {
            if (sender.Equals(treeView1))
            {
                e.Effect = DragDropEffects.Move;
            }
        }

        private void treeView1_DragOver(object sender, DragEventArgs e)
        {
            if ((e.KeyState & 8) == 8 && (e.AllowedEffect & DragDropEffects.Copy) == DragDropEffects.Copy)
            {
                e.Effect = DragDropEffects.Copy;
            }
            else if ((e.AllowedEffect & DragDropEffects.Move) == DragDropEffects.Move)
            {
                e.Effect = DragDropEffects.Move;
            }

            TreeView tv = sender as TreeView; 
            Point pt = tv.PointToClient(new Point(e.X, e.Y));

            int delta = tv.Height - pt.Y;
            if ((delta < tv.Height / 2) && (delta > 0))
            {
                TreeNode tn = tv.GetNodeAt(pt.X, pt.Y);
                if (tn != null && tn.NextVisibleNode != null)
                    tn.NextVisibleNode.EnsureVisible();
            }
            if ((delta > tv.Height / 2) && (delta < tv.Height))
            {
                TreeNode tn = tv.GetNodeAt(pt.X, pt.Y);
                if (tn != null && tn.PrevVisibleNode != null)
                    tn.PrevVisibleNode.EnsureVisible();
            } 
        }

        private void treeView1_BeforeLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
            if (e.Node.Level > 0)
            {
                e.CancelEdit = true;
                if (!RenamingTipShowed)
                {
                    MessageBox.Show("Only group names can be renamed. Channel names must be renamed from table.");
                    RenamingTipShowed = true;
                }
            }
        }

        private void deleteToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            //Delete cselected node
            if (treeView1.SelectedNode != null && treeView1.SelectedNode.Level == 0)
            {
                foreach (TreeNode tn in treeView1.SelectedNode.Nodes)
                {
                    DataRow dr = ChannelTable.menu.Tables["Menu"].Rows.Find(((ChanInfo)tn.Tag).Ip);
                    dr["Group"] = "";
                }

                constructTreeFromTable();
            }
        }

        private void addGroupToolStripMenuItem_Click(object sender, EventArgs e)
        {
            // add root node
            TreeNode tnn = new TreeNode();
            tnn.Text = "New Group";
            tnn.Name = "New Group";
            tnn.ImageIndex = 1;
            tnn.ToolTipText = "Click to rename";
            tnn.ContextMenuStrip = contextMenuStrip3;
            treeView1.Nodes.Add(tnn);

            if (!groupList.Contains("New Group"))
                groupList.Add("New Group");

            UpdateContextMenu();

        }

        private void treeView1_AfterLabelEdit(object sender, NodeLabelEditEventArgs e)
        {
            if (e.Node.Name == "" || e.Label ==null || e.Label == string.Empty || e.Label.Contains(","))
                return;

            string prevgroup = e.Node.Name;
            string newgroup = e.Label;
            e.Node.Name = e.Label;

            foreach (TreeNode tn in e.Node.Nodes)
            {
                UpdateBelongingGroups((ChanInfo)tn.Tag, prevgroup, newgroup,false);
            }

            if (!groupList.Contains(e.Label))
                groupList.Add(e.Label);
            if (groupList.Contains(prevgroup))
                groupList.Remove(prevgroup);

            // Update datatable
            updateTableFromTree(e.Node.Nodes);
            UpdateContextMenu();

        }

        private void removeAllGroupsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Remove all groups?", "", MessageBoxButtons.OKCancel) == DialogResult.OK)
            {
                foreach (DataRow dr in ChannelTable.menu.Tables["Menu"].Rows)
                {
                    dr["Group"] = "";
                }

                constructTreeFromTable();
            }
           
        }

        private void UpdateBelongingGroups(ChanInfo ci, string oldgroup, string newgroup, bool copy)
        {
            string[] groupes = ci.Group.Split(",".ToCharArray());
            string newstring="";
            bool allreadyBelongsToNewGroup = false;

            if (newgroup == "")
            {
                ci.Group = "";
                return;
            }

            int i = 0;
            foreach (string group in groupes)
            {
                if (group == newgroup)
                {
                    allreadyBelongsToNewGroup = true;
                    break;
                }
                if (group == oldgroup && !copy) // Rename group when moving
                {
                    groupes[i] = newgroup;
                }
                i += 1;
            }
             // convert back to string
            foreach (string group in groupes)
            {
                newstring += group + ",";
            }
            
            if (copy && !allreadyBelongsToNewGroup) // Add group when cloning
                newstring += newgroup;

            newstring = newstring.Trim(",".ToCharArray());
            ci.Group = newstring;

        }

        void groupMenuItem_Click(object sender, EventArgs e)
        {
            ToolStripItem tsi = (ToolStripItem)sender;
            string group = tsi.Text;

            TreeNode selectedNode = treeView1.SelectedNode;
            TreeNode[] destRootNode = treeView1.Nodes.Find(group, false);
            if (selectedNode != null && selectedNode.Level >0 && destRootNode.Length>0)
            {
                tnToMove = selectedNode;
                MoveItem(false, destRootNode[0]);
            }
        }

        private void UpdateContextMenu()
        {
            contextMenuStrip4.Items.Clear();

            ToolStripItem header = contextMenuStrip4.Items.Add("Send to group:");
            header.Font = new Font(header.Font, FontStyle.Bold);
            contextMenuStrip4.Items.Add("-");

            foreach (string group in groupList)
            {
                ToolStripItem item = contextMenuStrip4.Items.Add(group);
                item.Click += new EventHandler(groupMenuItem_Click);
            }
        }


        private void treeView1_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
                treeView1.SelectedNode = e.Node;

            if (e.Button == MouseButtons.Left && e.Node.Level>0)
            {
                foreach (DataGridViewRow dgvr in dataGridView1.Rows)
                {
                    if (!dgvr.IsNewRow)
                    {
                        DataRowView DataRowView = (DataRowView)dgvr.DataBoundItem;
                        DataRow row = DataRowView.Row;
                        if ((string)row["Ip"] == ((ChanInfo)e.Node.Tag).Ip)
                        {
                            dataGridView1.CurrentCell = dgvr.Cells[0];
                            break;
                        }
                    }
                }
            }
        }

        private void dataGridView1_CellEndEdit_1(object sender, DataGridViewCellEventArgs e)
        {

            if (e.ColumnIndex <= 3 && e.RowIndex >= 0)
            {
                try
                {
                    bindingSource1.EndEdit();
                }
                catch { };
                constructTreeFromTable();
            }
            if (e.ColumnIndex == 3 && e.RowIndex >= 0)
                dataGridView1.Rows[e.RowIndex].Cells[3].Value = ((string)dataGridView1.Rows[e.RowIndex].Cells[3].Value).TrimEnd(",".ToCharArray());

        }

        private static void setLanguage()
        {
             language = Thread.CurrentThread.CurrentCulture.IetfLanguageTag;
             Thread.CurrentThread.CurrentUICulture = new CultureInfo(Thread.CurrentThread.CurrentCulture.IetfLanguageTag);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }


        //=====================================================================================================
        // Scanner
        //=====================================================================================================

        const int BASIC_DATA_TIMEOUT= 5000; // 5seconds PAT & PMTs

        volatile IPAddress m_nextAddress;
        volatile IPAddress m_stopAddress;

        Thread[] m_threads = { null };
        Scanner.Scanner[] m_scanners = { null };
        
        bool m_scanners_initialized=false;
        bool m_WaitForSDT = true;
         
        int m_nextChanNumber=1;
        
        volatile int m_waitTimeout = 2000;
        volatile int m_numberOfThreads = 0;
        volatile bool m_markNewRows = false;
        volatile int m_exitCount = 0;
        volatile bool m_StopScanner = false;
        volatile uint m_NumOfAddresses = 0;   
     
        uint m_newCount = 0;
        uint m_scanCount = 0;

        public delegate void delegateVoidParseResult(ParseResult pr);
        public delegate void delegateVoidString(string str);
        public delegate void delegateVoidVoid();
        public delegate List<UdpSrcCom.ServiceDescriptorTriplet> delegateIntVoid();
        public delegate void delegateIpAddressUshort(IPAddress ip, ushort port);
        public delegate void delegateVoidInt(int percent);
        
        private List<ComboBoxItem> m_ComboBoxCollection;
        private NetworkInterfaces m_nics = new NetworkInterfaces();

        DataTable m_groupsTable = null;
        ManualResetEvent threadsStartedEvent = new ManualResetEvent(false);

        private void eventWait(object obj)
        {
            threadsStartedEvent.WaitOne();              
            
            Scanner.Scanner sc = (Scanner.Scanner)obj;
            
            const int E_ABORT = unchecked((int)0x80004004);
            int hr;
            IntPtr p1, p2;
            EventCode ec;
            DateTime DataStart = DateTime.Now;
            DateTime SearchStart = DateTime.Now;
            bool sdt_found = false;
            bool basic_data_parsed = false;
            bool changeIp = false;
            bool hasData = false;
            bool newChannel = true;
            
            int timeout = sc.m_Timeout;
           
            // if SDT waiting is enabled we wait for SDT notification or max 2 seconds before switching ip address. 
            // If SDT waiting is not enabled, we switch after EC_PARSER_COMPLETE message or after 2 seconds;
            do
            {
                lock (sc)
                {
                    sc.m_mre.WaitOne(100, true);
                    if (sc.m_State == Scanner.GraphState.Exiting || sc.m_State == Scanner.GraphState.NoGraph)
                        break;

                    for (
                        hr = sc.m_mediaEvent.GetEvent(out ec, out p1, out p2, 0);
                        hr >= 0;
                        hr = sc.m_mediaEvent.GetEvent(out ec, out p1, out p2, 0)
                        )
                    {

                        if ((int)ec == TSParser.EC_PARSER_STARTED) // Parser started
                        {
                            newChannel = true;
                            changeIp = false;
                            SearchStart = DateTime.Now;
                        }

                        else if ((int)ec == TSParser.EC_MSG_ParserRecvdFirstSample) // Parser received first sample.
                        {
                            hasData = true;
                            DataStart = DateTime.Now;
                        }

                        else if ((int)ec == TSParser.EC_PARSER_COMPLETE)    // Parser has found all data neccessary for creating directshow graph;
                        {
                            basic_data_parsed = true;
                        }

                        else if ((int)ec == TSParser.EC_MSG_SERVICE_LIST_RECEIVED) // Parser received service list.
                        {
                            sdt_found = true;
                        }
                        else
                        {
                            //don't process other messages
                        }
                        hr = sc.m_mediaEvent.FreeEventParams(ec, p1, p2);
                    }// end for
                }//endlock

                if (hr != E_ABORT)
                    DsError.ThrowExceptionForHR(hr);


                if (hasData)
                {
                    if (DateTime.Now.Subtract(DataStart).TotalMilliseconds > BASIC_DATA_TIMEOUT)
                        changeIp = true;
                    
                    if (m_WaitForSDT)
                    {
                        if (basic_data_parsed && sdt_found)
                        {
                            changeIp = true;
                        }
                    }
                    else
                    {
                        if (basic_data_parsed )
                            changeIp = true;
                    }
                }
                else // No data
                {
                    if (DateTime.Now.Subtract(SearchStart).TotalMilliseconds > timeout && newChannel)
                        changeIp = true;
                }

                if (changeIp)
                {
                    newChannel = false;

                    this.Invoke(new delegateVoidVoid(sc.stopFilter));                                // Stop filter.

                    ParseResult pr = new ParseResult(); 

                    lock (sc)
                    {
                        pr.IpPort = sc.m_multicastIp.ToString() + ":" + sc.m_multicastPort.ToString();

                        if (hasData)
                        {
                            if (sc.m_parser != null && !sc.m_parser.DataCollected)
                                this.Invoke(new delegateVoidVoid(sc.m_parser.GetInfo));              // Get parsed data.

                            if (sc.m_parser != null && sc.m_parser.ProgrameList.Count > 0)
                            {
                                pr.DataReceived = true;
                                List<UdpSrcCom.ServiceDescriptorTriplet> sdl = (List<UdpSrcCom.ServiceDescriptorTriplet>)this.Invoke(new delegateIntVoid(sc.m_parser.GetServiceDescriptor));
                                if (sdt_found && sdl.Count > 0)
                                    pr.Name = sdl[0].ServiceName;
                                if (sc.m_parser.ProgrameList[0].Audio.Count > 0)
                                    pr.FirstAudLang = sc.m_parser.ProgrameList[0].Audio[0].Language;
                            }
                        }
                    }

                    this.Invoke(new delegateVoidVoid(sc.cleanGraph));                           // Clean Graph.
                    dataGridView1.Invoke(new delegateVoidParseResult(UpdateDataTable), pr);    // Update datagridview.
                    label9.Invoke(new delegateVoidString(UpdateProgress), pr.IpPort);          // Update progress label
                    progressBar1.Invoke(new delegateVoidVoid(UpdateProgress));                 // Update progress bar    
                                                     
                    ushort port=0;
                    IPAddress ip = IPAddress.Any;
                    bool next = false;
                    bool finished = false;
                    
                    lock (this)
                    {
                        timeout = m_waitTimeout;                                                   // Refresh timeout value during scanning

                        if (IPAddressToLongBack(m_nextAddress) <= IPAddressToLongBack(m_stopAddress) && !m_StopScanner )
                        {
                            ip = m_nextAddress;
                            port = sc.m_multicastPort;
                            m_nextAddress = IncrementIP(m_nextAddress);
                            next = true;
                        }
                        else
                        {
                            m_exitCount += 1;
                            if (m_exitCount == Math.Min(m_NumOfAddresses,m_numberOfThreads))
                            {
                                m_markNewRows = false;// Last scanner finished
                                finished = true;
                            }
                        }
                    }

                    if (next)
                    {
                        this.Invoke(new delegateIpAddressUshort(sc.startGraph), ip, port);          // Start next read .
                        newChannel = true;
                    }
                    if (finished)
                    {
                        label9.Invoke(new delegateVoidString(UpdateProgress), "Finished");
                        button1.Invoke(new delegateVoidVoid(EnableScanButton));
                        if (treeView1.InvokeRequired)
                            treeView1.Invoke(new delegateVoidVoid(constructTreeFromTable));
                        else
                            constructTreeFromTable();
                    }
                    
                    //reset
                    changeIp = false;
                    hasData = false;
                    basic_data_parsed = false;
                    sdt_found = false;
                 }

            }
            while (true);

            sc.m_mreEventLoop.Set();
        }


        // START SCANNING
        private void button1_Click(object sender, EventArgs e)
        {
            IPAddress start;
            IPAddress stop;
            IPAddress NicIp = IPAddress.Any;
            int NicIndex = 0;
            m_exitCount = 0;
            m_newCount = 0;
            m_scanCount = 0;

            ushort port = (ushort)numericUpDown4.Value;
            m_waitTimeout = (int)numericUpDown2.Value;
            int numOfThreads = (int)numericUpDown3.Value;
            m_numberOfThreads = numOfThreads;

            if (checkBox1.CheckState == CheckState.Checked)
                m_WaitForSDT = true;
            else
                m_WaitForSDT = false;

           if(!( isMulticastAddress(textBox1.Text) && isMulticastAddress(textBox2.Text)) )
           {
               MessageBox.Show("Enter valid multicast ip address.");
               return;
           }

            // Reset background colors in existing table an find last available channel number.
            m_nextChanNumber = 1;
            foreach (DataGridViewRow dgvr in dataGridView1.Rows)
            {
                dgvr.DefaultCellStyle.BackColor = Color.White;

                if (!dgvr.IsNewRow)
                {
                    DataRowView DataRowView = (DataRowView)dgvr.DataBoundItem;
                    DataRow row = DataRowView.Row;
                    int ch = (int)row["Chan"];
                    if (ch >= m_nextChanNumber)
                    {
                        dgvr.DefaultCellStyle.BackColor = Color.White;
                        m_nextChanNumber = ch + 1;
                    }
                }
            }

            if (comboBox1.SelectedIndex != -1)
                NicIndex = m_nics.getIndexFromName(((ComboBoxItem)(comboBox1.SelectedItem)).Name);
            else
            {
                MessageBox.Show("Select NIC.");
                return;
            }

            // Open xml file with groups and audio languages, that will be later used.
            if (m_groupsTable == null)
            {
                m_groupsTable = new DataTable();
                m_groupsTable.Columns.Add("Name", typeof(string));
                m_groupsTable.Columns.Add("Group", typeof(string));
                m_groupsTable.Columns.Add("Audio", typeof(string));
                m_groupsTable.Columns["Name"].DefaultValue = "";
                m_groupsTable.Columns["Group"].DefaultValue = "";
                m_groupsTable.Columns["Audio"].DefaultValue = "";
                m_groupsTable.TableName = Thread.CurrentThread.CurrentCulture.IetfLanguageTag;
                m_groupsTable.PrimaryKey = new DataColumn[] { m_groupsTable.Columns["Name"] };
                try
                {
                    string xmlgroups = StartupPath + @"\parse\groups_" + Thread.CurrentThread.CurrentCulture.IetfLanguageTag + "_utf8.xml";
                    if (File.Exists(xmlgroups))
                        m_groupsTable.ReadXml(xmlgroups);
                    else
                        m_groupsTable.ReadXml(StartupPath + @"\parse\groups_default_utf8.xml");
                }
                catch { m_groupsTable = null; }
            }


            if (
                IPAddress.TryParse(textBox1.Text, out start) && 
                IPAddress.TryParse(textBox2.Text, out stop) &&
                IPAddress.TryParse(m_nics.getIpFromName(((ComboBoxItem)(comboBox1.SelectedItem)).Name), out NicIp)
                
                )
            {

                label9.Text = "Scanner Started";
                m_NumOfAddresses = calculateNumOfIpAddr(start, stop);
                progressBar1.Value = 0;
                m_StopScanner = false;
                button2.Enabled = true;

                if (IPAddressToLongBack(stop) < IPAddressToLongBack(start))
                {
                    m_stopAddress = start;
                    m_nextAddress = stop;
                }
                else
                {
                    m_stopAddress = stop;
                    m_nextAddress = start;
                }
               
                
                if (!m_scanners_initialized)
                {
                    m_threads = new System.Threading.Thread[numOfThreads];
                    m_scanners = new Scanner.Scanner[numOfThreads];

                    // initialize scanners
                    for (int i = 0; i < numOfThreads; i++)
                    {
                        m_scanners[i] = new Scanner.Scanner(NicIp, NicIndex, m_waitTimeout);
                        m_scanners[i].initGraph();
                        m_threads[i] = new Thread(eventWait);  // Start event thread
                        m_threads[i].Start(m_scanners[i]);
                    }

                    m_scanners_initialized = true;
                }

                for (int k = 0; k < numOfThreads; k++)
                {
                    IPAddress nextip=IPAddress.Any;
                    bool startGraph = false;

                        m_markNewRows = true;
                        if (IPAddressToLongBack(m_nextAddress) <= IPAddressToLongBack(m_stopAddress))
                        {
                            nextip = m_nextAddress;
                            m_nextAddress = IncrementIP(m_nextAddress);
                            startGraph = true;
                        }
                    
                    if (startGraph)
                        m_scanners[k].startGraph(nextip,port);
                }

                
                button1.Enabled = false;
                numericUpDown3.Enabled = false;
                threadsStartedEvent.Set();
            }
        }

        // STOP SCANNING
        private void button2_Click(object sender, EventArgs e)
        {
            m_StopScanner = true;
            //m_markNewRows = false;
            if (!label9.Text.StartsWith("Finished") && label9.Text!="" )
                label9.Text = "Stopping threads...";
            button2.Enabled = false;

       }

        private void UpdateDataTable (ParseResult pr)
        {
            m_scanCount += 1;

            DataRow dr = ChannelTable.menu.Tables["Menu"].Rows.Find(pr.IpPort);

            if (!pr.DataReceived )// No data received.
            {
                if (dr != null)
                {
                    // Iterate through datagridview and mark row with that ip address if exists.
                    foreach (DataGridViewRow dgvr in dataGridView1.Rows)
                    {
                        if (!dgvr.IsNewRow)
                        {
                            DataRowView DataRowView = (DataRowView)dgvr.DataBoundItem;
                            DataRow row = DataRowView.Row;
                            if ((string)row["Ip"] == pr.IpPort)
                            {
                                dgvr.DefaultCellStyle.BackColor = Color.LightSteelBlue;
                                break;
                            }
                        }
                    }
                }
            }
            else // Data received
            {
                if (dr == null) // New channel found
                {
                    DataRow menurow;
                    menurow = ChannelTable.menu.Tables["Menu"].NewRow();
                    DataRow drf = null; // entry in xml group table

                    m_newCount += 1;

                    int chan = m_nextChanNumber;
                    m_nextChanNumber += 1;
                    menurow["Ip"] = pr.IpPort;
                    menurow["Chan"] = chan;
                    if (pr.Name != "")
                    {
                        menurow["Name"] = pr.Name;
                        drf = m_groupsTable.Rows.Find(pr.Name);
                        if (drf!= null)
                            menurow["Group"] = (string)drf["Group"];
                    }
                    else
                        menurow["Name"] = "Channel " + chan.ToString();
                    if (pr.FirstAudLang != "")
                        menurow["Audio"] = pr.FirstAudLang;
                    else
                    {
                        // Try to get it from xml table
                        if (m_groupsTable != null)
                        {
                            if (drf != null)
                                menurow["Audio"] = (string)drf["Audio"];
                        }
                    }

                    menurow["EpgId"] = "";

                    ChannelTable.menu.Tables["Menu"].Rows.Add(menurow);
                }

                else 
                {
                    if (dr != null)
                    {
                        foreach (DataGridViewRow dgvr in dataGridView1.Rows)
                        {
                            if (!dgvr.IsNewRow)
                            {
                                DataRowView DataRowView = (DataRowView)dgvr.DataBoundItem;
                                DataRow row = DataRowView.Row;
                                if ((string)row["Ip"] == pr.IpPort)
                                {
                                    dgvr.DefaultCellStyle.BackColor = Color.Cornsilk;
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            //if (treeView1.InvokeRequired)
            //    treeView1.Invoke(new delegateVoidVoid(constructTreeFromTable));
            //else
            //    constructTreeFromTable();
       }

        private void UpdateProgress(string str)
        {
 
            label9.Text = str + "       New: " +m_newCount+ " ";
           
        }

        private void UpdateProgress()
        {

            if (m_NumOfAddresses>0)
                progressBar1.Value = (int)(m_scanCount * 100 / m_NumOfAddresses);
        }

        private void EnableScanButton()
        {
            button1.Enabled = true;
        }

        // helper functions
        static private uint IPAddressToLongBack(IPAddress IP)
        {
            byte[] byteIP = IP.GetAddressBytes();

            uint ip = (uint)byteIP[0] << 24;
            ip += (uint)byteIP[1] << 16;
            ip += (uint)byteIP[2] << 8;
            ip += (uint)byteIP[3];

            return ip;
        }

        static IPAddress IncrementIP(IPAddress addr)
        {
            byte[] ip = addr.GetAddressBytes();
            ip[3]++;
            if (ip[3] == 0)
            {
                ip[2]++;
                if (ip[2] == 0)
                {
                    ip[1]++;
                    if (ip[1] == 0)
                        ip[0]++;
                }
            }
            return new IPAddress(ip);
        }


        static uint calculateNumOfIpAddr(IPAddress start, IPAddress stop)
        {
            byte[] str = start.GetAddressBytes();
            byte[] stp = stop.GetAddressBytes();

            uint sta = (uint)255 * 255 * 255 * str[0] + (uint)255 * 255 * str[1] + (uint)255 * str[2] + (uint)str[3];
            uint stb = (uint)255 * 255 * 255 * stp[0] + (uint)255 * 255 * stp[1] + (uint)255 * stp[2] + (uint)stp[3];

            return (uint)Math.Abs((stb - sta)) + 1;
        }

        static bool isMulticastAddress(string ip)
        {
            if (Regex.IsMatch(ip, @"^\b(224|225|226|227|228|229|230|231|232|233|234|235|236|237|238|239).(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5]?)\b"))
                return true;
            else
                return false;
        }

        private void dataGridView1_RowsAdded(object sender, DataGridViewRowsAddedEventArgs e)
        {
            if (m_markNewRows)
            {
                dataGridView1.Rows[e.RowIndex].DefaultCellStyle.BackColor = Color.LightGreen;
            }
        }

        private void numericUpDown2_ValueChanged(object sender, EventArgs e)
        {
            lock (this)
            {
                m_waitTimeout = (int)numericUpDown2.Value;
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            lock (this)
            {
                for (int i = 0; i < m_numberOfThreads; i++)
                {
                    if (m_scanners[i] != null)
                        m_scanners[i].DestroyGraph();
                }
            }
            // save settings
            Properties.Settings.Default.WaitSdt = checkBox1.Checked;
            Properties.Settings.Default.WaitTimeout = (int)numericUpDown2.Value;
            Properties.Settings.Default.NumberOfThreads = (int)numericUpDown3.Value;
            Properties.Settings.Default.Port = (ushort)numericUpDown4.Value;
            Properties.Settings.Default.StartIp = textBox1.Text;
            Properties.Settings.Default.StopIp = textBox2.Text;
            if (comboBox1.Items.Count>0)
                 Properties.Settings.Default.NicIndex = comboBox1.SelectedIndex;

            Properties.Settings.Default.XspfExtension = toolStripComboBox1.SelectedIndex;

            Properties.Settings.Default.Save();

        }

        // Play 
        private void playToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (dataGridView1.SelectedRows.Count > 0)
            {
                string ipport = dataGridView1.SelectedRows[0].Cells[0].Value.ToString();
                string ip;
                string port;
                IPAddress iaIp=IPAddress.Any;
                ushort usPort=0;

                bool validaddress = false;
                if (ipport.Contains(":"))
                {
                    ip = ipport.Split(":".ToCharArray())[0];
                    port = ipport.Split(":".ToCharArray())[1];

                    if (isMulticastAddress(ip))
                    {
                        if (IPAddress.TryParse(ip, out iaIp) && ushort.TryParse(port,  out usPort))
                            validaddress = true;
                    }
                    else
                        MessageBox.Show("Not a multicast address", "Invalid Internet Protocol (IP) Address.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

                if (validaddress)
                {
                    IPAddress NicIp = IPAddress.Any;
                    int NicIndex = m_nics.getIndexFromName(((ComboBoxItem)(comboBox1.SelectedItem)).Name);
                    if (IPAddress.TryParse(m_nics.getIpFromName(((ComboBoxItem)(comboBox1.SelectedItem)).Name), out NicIp))
                    {
                        Player.PlayerWindow player = new Player.PlayerWindow(NicIp, NicIndex, usPort, iaIp);
                        player.Show();
                    }
                }
            }
        }

    }

    class ComboBoxItem
    {
        private string name;
        private int index;

        public ComboBoxItem(string name, int index)
        {
            this.name = name;
            this.index = index;
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public int Index
        {
            get { return index; }
            set { index = value; }
        }
    }

    public class ParseResult
    {
        private string ipPort="";
        private string name="";
        private string firstAudLang="";
        private bool dataReceived=false;

        public bool DataReceived
        {
            get { return dataReceived; }
            set { dataReceived = value;  }
        }
        
        public string IpPort
        {
            get { return ipPort; }
            set { ipPort = value; }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public string FirstAudLang
        {
            get { return firstAudLang; }
            set { firstAudLang = value; }
        }
    }

    public class ChanInfo
    {
        int chanNum;
        string group;
        string ip;
        string name;

        public ChanInfo(int num, string grp, string ipaddr,string ChanName)
        {
            chanNum = num;
            group = grp;
            ip = ipaddr;
            name = ChanName;
        }

        public int ChanNum
        {
            get { return chanNum; }
            set { chanNum = value; }
        }

        public string Group
        {
            get { return group; }
            set { group = value; }
        }

        public string Ip
        {
            get { return ip; }
            set { ip = value; }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }
    }
}
